@@ -19,6 +19,11 @@
  * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
+/*
+ * Modified 2010/09/12 by Colin Dean, Durham University IT Service,
+ * to add clustering support.
+ */
+
 #define	MAIN_PROGRAM
 
 #include <cron.h>
@@ -47,6 +52,13 @@ static int DisableInotify;
 
 #if defined WITH_INOTIFY
 
+/*
+ * Note that inotify isn't safe to use with clustering, as changes made
+ * to a shared filesystem on one system cannot be relied on to be notified
+ * on another system, so use of inotify is disabled at runtime if run with
+ * clustering enabled.
+ */
+
 # define NUM_WATCHES 3
 
 int wd[NUM_WATCHES];
@@ -105,7 +117,7 @@ static void handle_signals(cron_db * database) {
 		got_sighup = 0;
 #if defined WITH_INOTIFY
 		/* watches must be reinstated on reload */
-		if (inotify_enabled) {
+		if (inotify_enabled && (EnableClustering != 1)) {
 			set_cron_unwatched(database->ifd);
 			inotify_enabled = 0;
 		}
@@ -123,11 +135,18 @@ static void handle_signals(cron_db * database) {
 static void usage(void) {
 	const char **dflags;
 
-	fprintf(stderr, "usage:  %s [-n] [-p] [-s] [-i] [-m <mail command>] [-x [",
+	fprintf(stderr, "usage:  %s [-h] print this message \n \
+		[-i] deamon runs without inotify support \n \
+		[-m <mail command>] off or specify prefered client for sending mails \n \
+		[-n] run in foreground \n \
+		[-p] permit any crontab \n \
+		[-c] enable clustering support \n \
+		[-s] log into syslog instead of sending mails \n \
+		[-x [",
 		ProgramName);
 	for (dflags = DebugFlagNames; *dflags; dflags++)
 		fprintf(stderr, "%s%s", *dflags, dflags[1] ? "," : "]");
-	fprintf(stderr, "]\n");
+	fprintf(stderr, "] print debug information\n");
 	exit(ERROR_EXIT);
 }
 
@@ -220,6 +239,11 @@ int main(int argc, char *argv[]) {
 		}
 	}
 
+	if (access("/usr/sbin/sendmail", X_OK) != 0) {
+		SyslogOutput=1;
+		log_it("CRON", pid, "INFO","Syslog will be used instead of sendmail.", errno);
+	}
+
 	pid = getpid();
 	acquire_daemonlock(0);
 	database.head = NULL;
@@ -230,8 +254,8 @@ int main(int argc, char *argv[]) {
 
 	fd = -1;
 #if defined WITH_INOTIFY
-	if (DisableInotify) {
-		log_it("CRON", getpid(), "No inotify - daemon runs with -i option", 
+	if (DisableInotify || EnableClustering) {
+		log_it("CRON", getpid(), "No inotify - daemon runs with -i or -c option", 
 			"", 0);
 	}
 	else {
@@ -285,7 +309,7 @@ int main(int argc, char *argv[]) {
 			check_inotify_database(&database);
 		}
 		else {
-			if (load_database(&database))
+			if (load_database(&database) && (EnableClustering != 1))
 				/* try reinstating the watches */
 				set_cron_watched(fd);
 		}
@@ -383,7 +407,7 @@ int main(int argc, char *argv[]) {
 	}
 
 #if defined WITH_INOTIFY
-	if (inotify_enabled)
+	if (inotify_enabled && (EnableClustering != 1))
 		set_cron_unwatched(fd);
 
 	if (fd >= 0 && close(fd) < 0)
@@ -582,29 +606,34 @@ static void sigchld_reaper(void) {
 static void parse_args(int argc, char *argv[]) {
 	int argch;
 
-	while (-1 != (argch = getopt(argc, argv, "npsix:m:"))) {
+	while (-1 != (argch = getopt(argc, argv, "hnpsix:m:c"))) {
 		switch (argch) {
-		default:
-			usage();
-		case 'x':
-			if (!set_debug_flags(optarg))
+			case 'x':
+				if (!set_debug_flags(optarg))
+					usage();
+				break;
+			case 'n':
+				NoFork = 1;
+				break;
+			case 'p':
+				PermitAnyCrontab = 1;
+				break;
+			case 's':
+				SyslogOutput = 1;
+				break;
+			case 'i':
+				DisableInotify = 1;
+				break;
+			case 'm':
+				strncpy(MailCmd, optarg, MAX_COMMAND);
+				break;
+			case 'c':
+				EnableClustering = 1;
+				break;
+			case 'h':
+			default:
 				usage();
-			break;
-		case 'n':
-			NoFork = 1;
-			break;
-		case 'p':
-			PermitAnyCrontab = 1;
-			break;
-		case 's':
-			SyslogOutput = 1;
-			break;
-		case 'i':
-			DisableInotify = 1;
-			break;
-		case 'm':
-			strncpy(MailCmd, optarg, MAX_COMMAND);
-			break;
+				break;
 		}
 	}
 }
